/*
 * Copyright 1988 by the Massachusetts Institute of Technology.
 *
 * For copying and distribution information, please see the file
 * <mit-cpyright.h>.
 *
 * Declaration of debugging flag.
 */

#include <afsconfig.h>
#include <afs/param.h>


#include "mit-cpyright.h"
int des_debug = 0;
